#ifndef BLADERF2_COMPATIBILITY_H_
#define BLADERF2_COMPATIBILITY_H_

#include "helpers/version.h"

extern const struct version_compat_table bladerf2_fw_compat_table;
extern const struct version_compat_table bladerf2_fpga_compat_table;

#endif
